@@ -136,6 +136,7 @@ struct multiname_info
 	multiname* cached;
 	multiname_info():cached(NULL){}
 	~multiname_info(){delete cached;}
+	bool isAttributeName() const;
 };
 
 struct cpool_info
@@ -225,7 +226,6 @@ private:
 	u30 return_type;
 	std::vector<u30> param_type;
 	std::vector<option_detail> options;
-	u30 name;
 	u8 flags;
 
 	std::vector<u30> param_names;
@@ -265,8 +265,14 @@ private:
 	void doAnalysis(std::map<unsigned int,block_info>& blocks, llvm::IRBuilder<>& Builder);
 
 public:
+#ifdef PROFILING_SUPPORT
+	std::map<method_info*,uint64_t> profCalls;
+	uint64_t profTime;
+#endif
+
 	u30 option_count;
 	SyntheticFunction::synt_function f;
+	u30 name;
 	ABCContext* context;
 	method_body_info* body;
 	SyntheticFunction::synt_function synt_method();
@@ -275,7 +281,11 @@ public:
 	bool hasOptional() { return (flags & HAS_OPTIONAL) != 0;}
 	ASObject* getOptional(unsigned int i);
 	int numArgs() { return param_count; }
-	method_info():option_count(0),f(NULL),context(NULL),body(NULL)
+	method_info():
+#ifdef PROFILING_SUPPORT
+		profTime(0),
+#endif
+		option_count(0),f(NULL),context(NULL),body(NULL)
 	{
 	}
 };
@@ -423,6 +433,7 @@ public:
 	multiname* getMultiname(unsigned int m, call_context* th);
 	void buildInstanceTraits(ASObject* obj, int class_index);
 	ABCContext(std::istream& in) DLL_PUBLIC;
+	~ABCContext() DLL_PUBLIC;
 	void exec();
 
 	static bool isinstance(ASObject* obj, multiname* name);
@@ -452,10 +463,10 @@ private:
 	void registerFunctions();
 	//Interpreted AS instructions
 	static bool hasNext2(call_context* th, int n, int m); 
-	static void callPropVoid(call_context* th, int n, int m); 
-	static void callSuperVoid(call_context* th, int n, int m); 
-	static void callSuper(call_context* th, int n, int m); 
-	static void callProperty(call_context* th, int n, int m); 
+	static void callPropVoid(call_context* th, int n, int m, method_info*& called_mi);
+	static void callSuperVoid(call_context* th, int n, int m, method_info*& called_mi);
+	static void callSuper(call_context* th, int n, int m, method_info*& called_mi);
+	static void callProperty(call_context* th, int n, int m, method_info*& called_mi);
 	static void constructProp(call_context* th, int n, int m); 
 	static void setLocal(int n); 
 	static void setLocal_int(int n,int v); 
@@ -502,12 +513,15 @@ private:
 	static void pushUInt(call_context* th, int n);
 	static void pushDouble(call_context* th, int n);
 	static void incLocal_i(call_context* th, int n);
+	static void incLocal(call_context* th, int n);
+	static void decLocal_i(call_context* th, int n);
+	static void decLocal(call_context* th, int n);
 	static void coerce(call_context* th, int n);
 	static ASObject* getProperty(ASObject* obj, multiname* name);
 	static intptr_t getProperty_i(ASObject* obj, multiname* name);
 	static void setProperty(ASObject* value,ASObject* obj, multiname* name);
 	static void setProperty_i(intptr_t value,ASObject* obj, multiname* name);
-	static void call(call_context* th, int n);
+	static void call(call_context* th, int n, method_info*& called_mi);
 	static void constructSuper(call_context* th, int n);
 	static void construct(call_context* th, int n);
 	static void constructGenericType(call_context* th, int n);
@@ -526,14 +540,17 @@ private:
 	static bool _not(ASObject*);
 	static bool equals(ASObject*,ASObject*);
 	static number_t negate(ASObject*);
+	static intptr_t negate_i(ASObject*);
 	static void pop();
 	static ASObject* typeOf(ASObject*);
 	static void _throw(call_context* th);
+	static ASObject* asType(ASObject* obj, multiname* name);
 	static ASObject* asTypelate(ASObject* type, ASObject* obj);
 	static bool isTypelate(ASObject* type, ASObject* obj);
 	static bool isType(ASObject* obj, multiname* name);
 	static void swap();
 	static ASObject* add(ASObject*,ASObject*);
+	static intptr_t add_i(ASObject*,ASObject*);
 	static ASObject* add_oi(ASObject*,intptr_t);
 	static ASObject* add_od(ASObject*,number_t);
 	static uintptr_t bitAnd(ASObject*,ASObject*);
@@ -548,10 +565,12 @@ private:
 	static uintptr_t lShift(ASObject*,ASObject*);
 	static uintptr_t lShift_io(uintptr_t,ASObject*);
 	static number_t multiply(ASObject*,ASObject*);
+	static intptr_t multiply_i(ASObject*,ASObject*);
 	static number_t multiply_oi(ASObject*, intptr_t);
 	static number_t divide(ASObject*,ASObject*);
 	static intptr_t modulo(ASObject*,ASObject*);
 	static number_t subtract(ASObject*,ASObject*);
+	static intptr_t subtract_i(ASObject*,ASObject*);
 	static number_t subtract_oi(ASObject*, intptr_t);
 	static number_t subtract_io(intptr_t, ASObject*);
 	static number_t subtract_do(number_t, ASObject*);
@@ -565,6 +584,7 @@ private:
 	static intptr_t convert_i(ASObject*);
 	static uintptr_t convert_u(ASObject*);
 	static number_t convert_d(ASObject*);
+	static ASObject* convert_s(ASObject*);
 	static bool convert_b(ASObject*);
 	static bool greaterThan(ASObject*,ASObject*);
 	static bool greaterEquals(ASObject*,ASObject*);
@@ -611,6 +631,8 @@ private:
 	//It's sane to have them per-Vm, as anyway the vm is single by specs, single threaded
 	std::vector<thisAndLevel> method_this_stack;
 
+	//Profiling support
+	static uint64_t profilingCheckpoint(uint64_t& startTime);
 public:
 	GlobalObject* Global;
 	Manager* int_manager;
@@ -650,8 +672,6 @@ public:
 class DoABCTag: public ControlTag
 {
 private:
-	UI32_SWF Flags;
-	STRING Name;
 	ABCContext* context;
 	pthread_t thread;
 public:
@@ -660,6 +680,19 @@ public:
 	void execute(RootMovieClip* root);
 };
 
+class DoABCDefineTag: public ControlTag
+{
+private:
+	UI32_SWF Flags;
+	STRING Name;
+	ABCContext* context;
+	pthread_t thread;
+public:
+	DoABCDefineTag(RECORDHEADER h, std::istream& in);
+	~DoABCDefineTag();
+	void execute(RootMovieClip* root);
+};
+
 class SymbolClassTag: public ControlTag
 {
 private:
